#!/bin/ksh -r
#
#	$Id: cagair.ksh,v 1.70 2015/02/16 18:01:26 craski-shell_86 Exp $
#
# vim: tabstop=4 shiftwidth=4 softtabstop=4 noexpandtab
#
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
#
# Copyright (c) 2008 Christian Weisgerber. Public Domain.
#
# Copyright (c) 2014, 2015 Craig Skinner <skinner@britvault.co.uk>
#
# Permission to use, copy, modify, and distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
#


. ${LOCALBASE}/lib/sysexits.sh


thus=${0}
this=${thus##*/}
repo="${HOME}/.${this}"
that="${repo}/.${this}!${LOGNAME}"
conf="${that}.rc"
safe="${that}.txt.dat"
lock="${TMPDIR:-/tmp}/.${LOGNAME}.${this}.lock"
docs="${LOCALBASE}/${DOC_BASE}/${this}"
set -A editors 'mg' 'vi' 'vim'

export TMPDIR=${lock}
readonly this thus that repo conf safe lock docs editors TMPDIR


function bail_out
{
	[[ -d ${lock} ]] && rm -RP ${lock}
	[[ ${1} -eq 0 ]] && clear
	exit ${1}
}


function set_mode
{
	local _mode=${1}
	[[ -n ${_mode} && $# -gt 1 ]] ||
	{
		print -u2 "usage: $0() 6-digit-st_mode file/dir ..."
		bail_out ${EX_SOFTWARE}
	}

	shift
	for _item in $*
	do
		[[ -f ${_item} || -d ${_item} ]] || continue

		export $(stat -s ${_item})
		[[ ${st_mode} -eq ${_mode} ]] ||
		{
			chmod ${_mode} ${_item} || bail_out ${EX_NOPERM}
		}
	done
}


function dencrypt_file
{
	[[ $# -eq 3 ]] || return ${EX_SOFTWARE}
	[[ ${1} == '-d' || ${1} == '-e' ]] || return ${EX_SOFTWARE}
	[[ -f ${2} && -f ${3} ]] || return ${EX_IOERR}

	print -n -- "${password}" | \
		openssl aes-256-cbc ${1} -in ${2} -out ${3} \
				-pass stdin || return ${EX_DATAERR}
}


function get_editor
{
	# Remove any of the user's editor flags:
	unalias ${EDITOR} ${VISUAL}

	[[ -f ${conf} && -s ${conf} ]] && . ${conf}

	editor=${editor-${EDITOR-${VISUAL}}}
	[[ -n ${editor} ]] &&
	{
		for _editor in ${editors[*]}
		do
			[[ ${editor##*/} == ${_editor} ]] &&
			{
				whence -p ${_editor} &&
				{
					editor=${_editor}
					read choice?'Change editor? [y/N]: '
					[[ ${choice} == @(y|Y) ]] || return
					break
				}
			}
		done
		unset editor
	}

	print 'Secured file editors are:'
	for _editor in ${editors[*]}
	do
		print -n "${_editor}:\t"
		whence -p ${_editor}
	done

	until [[ -n ${editor} ]]
	do
		read choice?"Select from above list? [Default is ${editors[0]}]: "
		for _editor in ${editors[*]}
		do
			[[ ${choice:-${editors[0]}} == ${_editor} ]] &&
			{
				whence -p ${_editor} &&
				{
					editor=${_editor}
					break
				}
			}
		done
	done

	set_editor
}


function set_editor
{
	read choice?'Save selected editor for next time? [Y/n]: '
	[[ ${choice} == @(n|N) ]] ||
	{
		[[ -f ${conf} ]] || touch ${conf}

		grep -q ^editor= ${conf} ||
		{
			print "H\na\neditor='${editor}'\n.\nw" | ed -s ${conf}
			return
		}

		print "H\n,s/^editor=.*/editor='${editor}'/\n.\nw" | ed -s ${conf}
	}
}


function secure_editor
{
	case ${editor} in
		mg)
			# By default, mg sets 'make-backup-files' and
			# 'leave-tmpdir-backups', in ${TMPDIR} - OK
			# Users can set 'backup-to-home-directory' in ~/.mg
			# "-n	Turn off backup file generation."
			readonly editor_args='-n -f column-number-mode'
			;;
		vi)
			# Set vi's clear text file recovery directory to ${TMPDIR}
			readonly NEXINIT="set secure recdir=${TMPDIR}"
			export NEXINIT
			;;
		vim)
			# Store vim's temporary files in ${TMPDIR}
			VIMINIT="set secure directory=${TMPDIR}//"
			VIMINIT="${VIMINIT} backupdir=${TMPDIR}//"
			VIMINIT="${VIMINIT} undodir=${TMPDIR}//"
			VIMINIT="${VIMINIT} viminfo+=n${TMPDIR}//viminfo"
			readonly VIMINIT
			export VIMINIT
			# -i {viminfo} ... used to skip the use of the .viminfo file
			#
			# -n	No swap file will be used. Recovery
			#	after a crash will be impossible.
			#
			# -Z	Restricted mode.
			readonly editor_args='-n -i NONE -Z'
			;;
		*)
			print -u2 "Unable to secure editor '${editor}'"
			bail_out ${EX_CONFIG}
			;;
	esac
}


function get_password
{
	stty -echo
	IFS= read -r password?'Password: '
	stty echo
}


function get_password_checker
{
	login_class=$(userinfo ${LOGNAME} | awk '/^class/ { print $2 }')
	password_checker=$(getcap -f /etc/login.conf -s passwordcheck ${login_class:='default'})
	[[ -n ${password_checker} ]] &&
	{
		# Restricted shell; strip path
		password_checker=${password_checker##*/}
		# Ensure it (without args) is in our $PATH
		whence -p ${password_checker% *} || unset password_checker
	}
}


function check_password
{
	[[ -n ${password_checker} ]] &&
	{
		print -- "${password}" | ${password_checker} || unset password
	}

	[[ -n ${password} ]] &&
	{
		password_1="${password}"
		get_password
		[[ "${password}" == "${password_1}" ]] ||
		{
			print -u2 'They do not match.'
			unset password password_1
		}
	}
}


function usage
{
	_usage="$0:\t${this} [-hv] [-i file]\noptions:\n"
	_usage="${_usage}\t-h\t# this help\n"
	_usage="${_usage}\t-i file\t# import plain text\n"
	_usage="${_usage}\t-v\t# version number\n"
	_usage="${_usage}\nFor more help and Scottish Gaelic pronunciation,\n"
	_usage="${_usage}refer to the included README and MP3 files in:\n${docs}/"
	_print_args='--'
	[[ ${1} -eq 0 ]] || _print_args="-u2 ${_print_args}"
	print ${_print_args} ${_usage}
	exit ${1}
}


function check_import_file
{
	[[ -f ${1} ]] ||
	{
		print -u2 "Not a file to import: ${1}"
		return ${EX_USAGE}
	}
	[[ -r ${1} ]] ||
	{
		print -u2 "Can't read import file: ${1}"
		return ${EX_NOPERM}
	}
	[[ -s ${1} ]] ||
	{
		print -u2 "Empty import file: ${1}"
		return ${EX_DATAERR}
	}
	file ${1} | fgrep -q -w text ||
	{
		print -u2 "Not a text file to import: ${1}"
		return ${EX_DATAERR}
	}
}


function singleton
{
	mkdir ${lock} && return
	[[ -d ${lock} ]] || return ${EX_UNAVAILABLE}

	set -A pids $(pgrep -f -u ${LOGNAME} "$(head -n 1 ${thus} | cut -d! -f2) .*/${this}")
	[[ ${#pids[*]} -eq 1 && ${pids[0]} -eq $$ ]] ||
	{
		print -u2 "ERROR: ${this} already running!"

		for pid in ${pids[*]}
		do
			[[ ${pid} -eq $$ ]] && continue
			ps -o 'pid,start,time,state,logname' -p ${pid}
		done | sort -u

		return ${EX_TEMPFAIL}
	}
}


# Alter/check the shell
trap 'bail_out' HUP INT QUIT ABRT TERM STOP
umask 077
for builtin in 'print' '[[' '[' 'whence'
do
	type ${builtin} | egrep -q ' is a shell builtin$| is a reserved word$' ||
	{
		print -u2 '*** Security breach! ***'
		print -u2 "The '${builtin}' command is not a shell builtin."
		exit ${EX_UNAVAILABLE}
	}
done


while getopts "hi:v" opt
do
	case ${opt} in
		h)
			usage
			;;
		i)
			import_file=${OPTARG}
			check_import_file ${import_file} || exit
			;;
		v)
			print -n 'Version: '
			cat "${docs}/VERSION"
			exit
			;;
		\?)
			usage ${EX_USAGE}
			;;
	esac
done


singleton || exit
readonly unsafe=$(mktemp) || bail_out ${EX_CANTCREAT}
chflags -R nodump ${lock}


[[ -d ${repo} ]] ||
{
	mkdir ${repo} || bail_out ${EX_CANTCREAT}
}
set_mode 040300 ${repo}
set_mode 100600 ${safe} ${conf}


get_editor
secure_editor
get_password_checker


trap 'bail_out $?' ERR


if [[ -f ${safe} ]]
then
	print 'Decrypting existing secrets...'
	get_password
	dencrypt_file -d ${safe} ${unsafe}
else
	touch ${safe}
fi
unsafe_cksum_before_edit=$(cksum ${unsafe})

[[ -n ${import_file} ]] &&
{
	print "H\na\n\n# import of ${import_file} @ $(date)\n\n.\nw" | ed -s ${unsafe}
	cp -p ${unsafe} ${unsafe}~
	sed -n w${unsafe} ${unsafe}~ ${import_file}
	print -n '\nImport file was appended. Continue [ENTER]: '
	read
}

${editor} ${editor_args} ${unsafe}
[[ ! -s ${safe} && ! -s ${unsafe} ]] &&
{
	print 'Nothing to encrypt, exiting.'
	rm ${safe} ${unsafe}
	bail_out ${EX_OK}
}

unsafe_cksum_after_edit=$(cksum ${unsafe})
[[ ${unsafe_cksum_before_edit} == ${unsafe_cksum_after_edit} ]] &&
{
	print 'No modification, exiting.'
	bail_out ${EX_OK}
}

print 'Encrypting secrets...'
[[ -n ${password} ]] &&
{
	read choice?'Reuse the same password again? [Y/n]: '
	[[ ${choice} == @(n|N) ]] && unset password
}
[[ -n ${password} ]] ||
{
	warning='\n*** WARNING!!! ***\n'
	warning="${warning}There is _ABSOLUTELY_NO_ password recovery procedure!\n"
	warning="${warning}*** Be VERY careful to REMEMBER this! ***\n"
	print ${warning}
	until [[ -n ${password} ]]
	do
		get_password
		check_password
	done
}
dencrypt_file -e ${unsafe} ${safe}
bail_out

